from unittest import mock
import farmbot as fb
import json

fake_bytes = b"{\"token\": {\"unencoded\": {\"aud\": \"unknown\", \"sub\": 1, \"iat\": 1609876696, \"jti\": \"46694c2f-2a03-4e3f-84bd-7ec22fb58c1f\", \"iss\": \"//10.11.1.235:3000\", \"exp\": 1613332696, \"mqtt\": \"10.11.1.235\", \"bot\": \"device_2\", \"vhost\": \"/\", \"mqtt_ws\": \"ws://10.11.1.235:3002/ws\"}, \"encoded\": \"eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJhdWQiOiJ1bmtub3duIiwic3ViIjoxLCJpYXQiOjE2MDk4NzY2OTYsImp0aSI6IjQ2Njk0YzJmLTJhMDMtNGUzZi04NGJkLTdlYzIyZmI1OGMxZiIsImlzcyI6Ii8vMTAuMTEuMS4yMzU6MzAwMCIsImV4cCI6MTYxMzMzMjY5NiwibXF0dCI6IjEwLjExLjEuMjM1IiwiYm90IjoiZGV2aWNlXzIiLCJ2aG9zdCI6Ii8iLCJtcXR0X3dzIjoid3M6Ly8xMC4xMS4xLjIzNTozMDAyL3dzIn0.mdiWPVn0Vp3tYbZfP9P2CRpKd3txjXkfUjvkgALeXAC7OJGSBu_0uIU3ZRakhfu6QaEKX0pP-jKlxvxgNDbYNgwUoqEC8RLrbnGoCx_SmhSf_43TRW2UoKOSo8BFp2D7_0L_zkocW_6VCzwLKo90m09N2VSA4S3wSItRphimEuvK224VSnr5oqyqQ5NfVgkA_usLVpExX6t7-5rF6Lm_VpIm6y-7Z4k6GvaJLH8WIDBsLQrM7STeFNwZGxWuUeblHjY4eyv2DkX7m1s9EN-fCUQzgBzvFUcjemk8rhAfz2A02SqShD2YHa2qidPnjpiJpwXWK9vXYdO2KQsjrsdBrA\"}, \"user\": {\"id\": 1, \"device_id\": 2, \"name\": \"Test\", \"email\": \"test@test.com\", \"created_at\": \"2020-12-14T20: 22: 49.821Z\", \"updated_at\": \"2021-01-05T19: 50: 12.653Z\", \"agreed_to_terms_at\": \"2020-12-14T20: 22: 49.854Z\", \"inactivity_warning_sent_at\": null}}"
fake_token = json.dumps({
    "token": {
        "unencoded": {
            "bot": "456",
            "exp": 1000,
            "iat": 900,
            "iss": "?",
            "jti": "Long string",
            "mqtt": "mqtt://my.farm.bot:1883",
            "mqtt_ws": "wss://my.farm.bot:1883",
            "sub": "device_456",
            "vhost": "dfsdfxcsd",
        },
        "encoded": "TOPSECRETTOKEN"
    }
})


class FakeReturn:
    def read(_) -> bytes:
        return fake_bytes


class FakeFarmbot(fb.Farmbot):
    def __init__(self) -> None:
        self.password = "drowssap"
        self.username = "emanresu"
        self.hostname = "tob.mraf.ym"
        self.state = None
        self.handle_connect = mock.MagicMock()
        self.handle_message = mock.MagicMock()
        self._handler = fb.StubHandler()
        self.read_status = mock.MagicMock()


class FakeMQTT():
    def __init__(self) -> None:
        self.connect = mock.MagicMock()
        self.loop_forever = mock.MagicMock()
        self.username_pw_set = mock.MagicMock()
        self.subscribe = mock.MagicMock()


class FakeMqttMessage():
    def __init__(self, topic: str, payload: str) -> None:
        self.topic = topic
        self.payload = payload


class TestFarmbotToken():
    @mock.patch("farmbot.urlopen", autospec=True, return_value=FakeReturn())
    def test_download_token(_, mock_download_token):
        email = "test@test.com"
        password = "password123"
        server = "http://localhost:3000"
        tkn = fb.FarmbotToken.download_token(email=email,
                                             password=password,
                                             server=server)
        assert isinstance(tkn, bytes)
        assert tkn == fake_bytes

    def test_init(_):
        token = fb.FarmbotToken(fake_bytes)
        assert(token.bot == "device_2")
        assert(token.exp == 1613332696)
        assert(token.iat == 1609876696)
        assert(token.iss == "//10.11.1.235:3000")
        assert(token.jti == "46694c2f-2a03-4e3f-84bd-7ec22fb58c1f")
        assert(token.mqtt_ws == "ws://10.11.1.235:3002/ws")
        assert(token.mqtt == "10.11.1.235")
        assert(token.sub == 1)
        assert(token.vhost == "/")
        assert(len(token.jwt) == 671)


class TestEmptyState():
    def test_empty_state(self):
        empty_xyz = {"x": None, "y": None, "z": None}
        state = fb.empty_state()
        assert(state["configuration"] == {})
        assert(state["informational_settings"] == {})
        assert(state["jobs"] == {})
        assert(state["location_data"]["axis_states"] == empty_xyz)
        assert(state["location_data"]["load"] == empty_xyz)
        assert(state["location_data"]["position"] == empty_xyz)
        assert(state["location_data"]["raw_encoders"] == empty_xyz)
        assert(state["location_data"]["scaled_encoders"] == empty_xyz)
        assert(state["mcu_params"] == {})
        assert(state["pins"] == {})
        assert(state["user_env"] == {})


class TestErrorResponse():
    def test_error_response(self):
        id = "my_id"
        errors = ["error1", "error2"]
        fake = fb.ErrorResponse(id, errors)
        assert(fake.id == id)
        assert(fake.errors == errors)


class TestOkResponse():
    def test_error_response(self):
        id = "my_id"
        fake = fb.OkResponse(id)
        assert(fake.id == id)


class TestFarmbotConnection():
    def test_init(self):
        my_farmbot = FakeFarmbot()
        connection = fb.FarmbotConnection(my_farmbot)
        assert(connection.bot == my_farmbot)
        assert(connection.mqtt)
        assert(connection.incoming_chan == "bot/emanresu/from_device")
        assert(connection.logs_chan == "bot/emanresu/logs")
        assert(connection.outgoing_chan == "bot/emanresu/from_clients")
        assert(connection.status_chan == "bot/emanresu/status")
        expected = ('bot/emanresu/status',
                    'bot/emanresu/logs',
                    'bot/emanresu/from_device')
        assert(connection.channels == expected)

    def test_start_connection(self):
        my_farmbot = FakeFarmbot()
        client = FakeMQTT()
        connection = fb.FarmbotConnection(my_farmbot, client)
        connection.start_connection()
        client.connect.assert_called_with('tob.mraf.ym', 1883, 60)
        client.loop_forever.assert_called()
        client.username_pw_set.assert_called_with('emanresu', 'drowssap')
        assert(client.on_connect)
        assert(client.on_message)

    def test_handle_connect(self):
        # https://docs.python.org/3/library/unittest.mock.html
        my_farmbot = FakeFarmbot()
        my_farmbot._handler.on_connect = mock.MagicMock()
        client = FakeMQTT()
        connection = fb.FarmbotConnection(my_farmbot, client)
        connection.handle_connect(mqtt=client,
                                  userdata=None,
                                  flags=None,
                                  rc=None)
        for channel in connection.channels:
            client.subscribe.assert_has_calls([mock.call(channel)])
        my_farmbot._handler.on_connect.assert_called_with(my_farmbot, client)
        my_farmbot.read_status.assert_called()

    def test_handle_message(self):
        my_farmbot = FakeFarmbot()
        conn = fb.FarmbotConnection(my_farmbot, FakeMQTT())

        # == STATUS Message
        conn.handle_status = mock.MagicMock()
        status_msg = FakeMqttMessage(conn.status_chan, '{}')
        conn.handle_message(conn.mqtt, None, status_msg)
        conn.handle_status.assert_called_with(status_msg)

        # == LOG Message
        conn.handle_log = mock.MagicMock()
        log_msg = FakeMqttMessage(conn.logs_chan, '{}')
        conn.handle_message(conn.mqtt, None, log_msg)
        conn.handle_log.assert_called_with(log_msg)

        # == RPC Message
        conn.unpack_response = mock.MagicMock()
        rpc_json = '{}'
        rpc_msg = FakeMqttMessage(conn.incoming_chan, rpc_json)
        conn.handle_message(conn.mqtt, None, rpc_msg)
        conn.unpack_response.assert_called_with(rpc_json)

    def test_unpack_response(self):
        conn = fb.FarmbotConnection(FakeFarmbot(), FakeMQTT())
        # == OK Response
        conn.handle_resp = mock.MagicMock()
        rpc_ok = json.dumps({"kind": "rpc_ok", "args": {"label": "123"}})
        conn.unpack_response(rpc_ok)
        conn.handle_resp.assert_called_with("123")

        # == ERROR Response
        conn.handle_error = mock.MagicMock()
        errors = [
            {'kind': 'explanation', 'args': {'message': 'ERROR 1'}},
            {'kind': 'explanation', 'args': {'message': 'ERROR 2'}},
        ]
        rpc_err = json.dumps({
            "kind": "rpc_error",
            "args": {"label": "456"},
            "body": errors
        })

        conn.unpack_response(rpc_err)
        conn.handle_error.assert_called_with("456", errors)

    def test_handle_resp(self):
        conn = fb.FarmbotConnection(FakeFarmbot(), FakeMQTT())
        my_mock = mock.MagicMock()
        conn.bot._handler.on_response = my_mock
        conn.handle_resp("any_label")
        args = my_mock.call_args[0]
        bot = args[0]
        response = args[1]
        assert bot == conn.bot
        assert response.id == "any_label"

    def test_handle_status(self):
        conn = fb.FarmbotConnection(FakeFarmbot(), FakeMQTT())
        status_msg = FakeMqttMessage(conn.status_chan, '{}')
        conn.bot._handler.on_change = mock.MagicMock()
        conn.handle_status(status_msg)
        conn.bot._handler.on_change.assert_called_with(conn.bot, {})

    def test_handle_log(self):
        conn = fb.FarmbotConnection(FakeFarmbot(), FakeMQTT())
        status_msg = FakeMqttMessage(conn.logs_chan, '{}')
        conn.bot._handler.on_log = mock.MagicMock()
        conn.handle_log(status_msg)
        conn.bot._handler.on_log.assert_called_with(conn.bot, {})

    def test_handle_error(self):
        errors = [
            {'kind': 'explanation', 'args': {'message': 'ERROR 1'}},
            {'kind': 'explanation', 'args': {'message': 'ERROR 2'}},
        ]
        label = 'fd0ee7c8-6ca8-11eb-9d9d-eba70539ce61'

        conn = fb.FarmbotConnection(FakeFarmbot(), FakeMQTT())
        status_msg = FakeMqttMessage(conn.incoming_chan, '{}')
        conn.bot._handler.on_error = mm = mock.MagicMock()
        conn.handle_error(label, errors)
        args = mm.call_args[0]
        actual_bot = args[0]
        actual_response = args[1]
        assert actual_bot == conn.bot
        assert actual_response.id == label
        assert actual_response.errors == ['ERROR 1', 'ERROR 2']

    @mock.patch("farmbot.uuid.uuid1", autospec=True, return_value="FAKE_UUID")
    def test_send_rpc(self, _):
        mqtt = FakeMQTT()
        mqtt.publish = mock.MagicMock()
        conn = fb.FarmbotConnection(FakeFarmbot(), mqtt)
        # === NON-ARRAY
        result = conn.send_rpc({})
        assert result == "FAKE_UUID"
        expected_chan = 'bot/emanresu/from_clients'
        expected_json = '{"kind": "rpc_request", "args": {"label": "FAKE_UUID"}, "body": [{}]}'
        mqtt.publish.assert_called_with(expected_chan, expected_json)
        # === ARRAY
        result2 = conn.send_rpc([{}])
        assert result2 == "FAKE_UUID"
        expected_chan2 = 'bot/emanresu/from_clients'
        expected_json2 = '{"kind": "rpc_request", "args": {"label": "FAKE_UUID"}, "body": [{}]}'
        mqtt.publish.assert_called_with(expected_chan2, expected_json2)

    fake_token = json.dumps({
        "token": {
            "unencoded": {
                "bot": "456",
                "exp": 1000,
                "iat": 900,
                "iss": "?",
                "jti": "Long string",
                "mqtt": "mqtt://my.farm.bot:1883",
                "mqtt_ws": "wss://my.farm.bot:1883",
                "sub": "device_456",
                "vhost": "dfsdfxcsd",
            },
            "encoded": "TOPSECRETTOKEN"
        }
    })

    @mock.patch("farmbot.FarmbotToken.download_token", autospec=True, return_value=fake_token)
    def test_login(self, dl_token):
        bot = fb.Farmbot.login("email", "pass")
        dl_token.assert_called_with(email='email',
                                    password='pass',
                                    server='https://my.farm.bot')
        assert bot.username == "456"
        assert bot.password == "TOPSECRETTOKEN"
        assert bot.hostname == "mqtt://my.farm.bot:1883"
        assert bot.device_id == "device_456"
        assert isinstance(bot._handler, fb.StubHandler)

        assert isinstance(bot._connection, fb.FarmbotConnection)
        assert bot.state == fb.empty_state()

    def test_position(self):
        bot = fb.Farmbot(fake_token)
        assert bot.position() == (-0.0, -0.0, -0.0)

    def test__do_cs(self):
        fake_rpc = {"kind": "nothing", "args": {}, "body": []}
        bot = fb.Farmbot(fake_token)
        bot._connection.send_rpc = mock.MagicMock()
        bot._do_cs("nothing", {})
        bot._connection.send_rpc.assert_called_with(fake_rpc)

    def test_connect(self):
        bot = fb.Farmbot(fake_token)
        fake_handler = 'In real life, this would be a class'
        bot._connection.start_connection = mock.MagicMock()
        bot.connect(fake_handler)
        assert bot._handler == fake_handler
        bot._connection.start_connection.assert_called()

    def expected_rpc(self, bot, kind, args, body=None):
        if not body:
            bot._do_cs.assert_called_with(kind, args)
        else:
            bot._do_cs.assert_called_with(kind, args, body)

    def test_various_rpcs(self):
        bot = fb.Farmbot(fake_token)
        bot._do_cs = mock.MagicMock()

        bot.send_message("Hello, world!")
        self.expected_rpc(bot,
                          'send_message',
                          {'message': 'Hello, world!', 'message_type': 'info'})

        bot.emergency_lock()
        self.expected_rpc(bot, "emergency_lock", {})

        bot.emergency_unlock()
        self.expected_rpc(bot, "emergency_unlock", {})

        bot.find_home()
        self.expected_rpc(bot, "find_home", {"speed": 100, "axis": "all"})

        bot.find_length()
        self.expected_rpc(bot, "calibrate", {"axis": "all"})

        bot.flash_farmduino("express_k10")
        self.expected_rpc(bot, "flash_firmware", {"package": "express_k10"})

        bot.go_to_home()
        self.expected_rpc(bot, "home", {"speed": 100, "axis": "all"})

        bot.move_absolute(x=1.2, y=3.4, z=5.6)
        self.expected_rpc(bot,
                          'move_absolute',
                          {
                              'location': {
                                  'kind': 'coordinate',
                                  'args': {'x': 1.2, 'y': 3.4, 'z': 5.6}
                              },
                              'speed': 100.0,
                              'offset': {
                                  'kind': 'coordinate',
                                  'args': {'x': 0, 'y': 0, 'z': 0}
                              }
                          })

        bot.move_relative(x=1.2, y=3.4, z=5.6)
        self.expected_rpc(bot,
                          "move_relative",
                          {"x": 1.2, "y": 3.4, "z": 5.6, "speed": 100})

        bot.power_off()
        self.expected_rpc(bot, "power_off", {})

        bot.read_status()
        self.expected_rpc(bot, "read_status", {})

        bot.reboot()
        self.expected_rpc(bot, "reboot", {"package": "farmbot_os"})

        bot.reboot_farmduino()
        self.expected_rpc(bot,
                          "reboot",
                          {"package": "arduino_firmware"})

        bot.factory_reset()
        self.expected_rpc(bot, "factory_reset", {"package": "farmbot_os"})

        bot.sync()
        self.expected_rpc(bot, "sync", {})

        bot.take_photo()
        self.expected_rpc(bot, "take_photo", {})

        bot.toggle_pin(12)
        self.expected_rpc(bot, "toggle_pin", {"pin_number": 12})

        bot.update_farmbot_os()
        self.expected_rpc(bot, "check_updates", {"package": "farmbot_os"})

        bot.read_pin(21)
        self.expected_rpc(bot,
                          "read_pin",
                          {'label': 'pin21', 'pin_mode': 0, 'pin_number': 21})

        bot.write_pin(45, 67)
        self.expected_rpc(bot,
                          "write_pin",
                          {'pin_mode': 0, 'pin_number': 45, 'pin_value': 67})

        bot.set_servo_angle(5, 90)
        self.expected_rpc(bot,
                          "set_servo_angle",
                          {'pin_number': 5, 'pin_value': 90})



import paho.mqtt.client as mqtt
from urllib.request import urlopen, Request
import json
import uuid


class OkResponse():
    def __init__(self, id):
        self.id = id


class ErrorResponse():
    def __init__(self, id, errors):
        self.errors = errors
        self.id = id


class FarmbotConnection():
    def __init__(self, bot, mqtt=mqtt.Client()):
        self.bot = bot
        self.mqtt = mqtt
        u = bot.username
        self.mqtt.username_pw_set(u, bot.password)
        # bot/device_000/from_clients
        # bot/device_000/from_device
        # bot/device_000/logs
        # bot/device_000/status
        # bot/device_000/sync/#
        # bot/device_000/ping/#
        # bot/device_000/pong/#
        self.status_chan = "bot/" + u + "/status"
        self.logs_chan = "bot/" + u + "/logs"
        self.incoming_chan = "bot/" + u + "/from_device"
        self.outgoing_chan = "bot/" + u + "/from_clients"
        self.channels = (
            self.status_chan,
            self.logs_chan,
            self.incoming_chan
        )

    def start_connection(self):
        # Attach event handlers:
        self.mqtt.on_connect = self.handle_connect
        self.mqtt.on_message = self.handle_message

        # Finally, connect to the server:
        self.mqtt.connect(self.bot.hostname, 1883, 60)

        self.mqtt.loop_forever()

    def handle_connect(self, mqtt, userdata, flags, rc):
        for channel in self.channels:
            mqtt.subscribe(channel)
        self.bot.read_status()
        self.bot._handler.on_connect(self.bot, mqtt)

    def handle_message(self, mqtt, userdata, msg):
        if msg.topic == self.status_chan:
            self.handle_status(msg)

        if msg.topic == self.logs_chan:
            self.handle_log(msg)

        if msg.topic == self.incoming_chan:
            self.unpack_response(msg.payload)

    def unpack_response(self, payload):
        resp = json.loads(payload)
        kind = resp["kind"]
        label = resp["args"]["label"]
        if kind == "rpc_ok":
            self.handle_resp(label)
        if kind == "rpc_error":
            self.handle_error(label, resp["body"] or [])

    def handle_resp(self, label):
        # {
        #   'kind': 'rpc_ok',
        #   'args': { 'label': 'fd0ee7c9-6ca8-11eb-9d9d-eba70539ce61' },
        # }
        self.bot._handler.on_response(self.bot, OkResponse(label))
        return

    def handle_status(self, msg):
        self.bot.state = json.loads(msg.payload)
        self.bot._handler.on_change(self.bot, self.bot.state)
        return

    def handle_log(self, msg):
        log = json.loads(msg.payload)
        self.bot._handler.on_log(self.bot, log)
        return

    def handle_error(self, label, errors):
        # {
        #   'kind': 'rpc_error',
        #   'args': { 'label': 'fd0ee7c8-6ca8-11eb-9d9d-eba70539ce61' },
        #   'body': [
        #       { 'kind': 'explanation', 'args': { 'message': '~ERROR~' } },
        #       { 'kind': 'explanation', 'args': { 'message': '~ERROR~' } },
        #     ],
        # }

        tidy_errors = []
        for error in errors:
            args = error["args"] or {"message": "No message provided"}
            message = args["message"]
            tidy_errors.append(message)
        self.bot._handler.on_error(self.bot, ErrorResponse(label, tidy_errors))
        return

    def send_rpc(self, rpc):
        label = str(uuid.uuid1())
        message = {"kind": "rpc_request", "args": {"label": label}}
        if isinstance(rpc, list):
            message["body"] = rpc
        else:
            message["body"] = [rpc]
        payload = json.dumps(message)
        self.mqtt.publish(self.outgoing_chan, payload)
        return label


class StubHandler:
    def on_connect(self, bot, client): pass
    def on_change(self, bot, state): pass
    def on_log(self, _bot, log): pass
    def on_error(self, _bot, _response): pass
    def on_response(self, _bot, _response): pass


class FarmbotToken():
    @staticmethod
    def download_token(email,
                       password,
                       server="https://my.farm.bot"):
        """
        Returns a byte stream representation of the
        """
        req = Request(server + "/api/tokens")
        req.add_header("Content-Type", "application/json")
        data = {"user": {"email": email, "password": password}}
        string = json.dumps(data)
        as_bytes = str.encode(string)
        response = urlopen(req, as_bytes)

        return response.read()

    def __init__(self, raw_token):
        token_data = json.loads(raw_token)
        self.raw_token = raw_token
        token = token_data["token"]["unencoded"]
        self.jwt = token_data["token"]["encoded"]

        self.bot = token["bot"]
        self.exp = token["exp"]
        self.iat = token["iat"]
        self.iss = token["iss"]
        self.jti = token["jti"]
        self.mqtt = token["mqtt"]
        self.mqtt_ws = token["mqtt_ws"]
        self.sub = token["sub"]
        self.vhost = token["vhost"]


empty_xyz = {"x": None, "y": None, "z": None}
zero_xyz = {"x": 0, "y": 0, "z": 0}


def empty_state():
    """
    This is what the bot's state tree looks like.
    This helper method creates an empty object that is mostly
    unpopulated.
    """
    return {
        "configuration": {},
        "informational_settings": {},
        "jobs": {},
        "location_data": {
            "axis_states": empty_xyz,
            "load": empty_xyz,
            "position": empty_xyz,
            "raw_encoders": empty_xyz,
            "scaled_encoders": empty_xyz,
        },
        "mcu_params": {},
        "pins": {},
        "user_env": {}
    }


class Farmbot():
    @classmethod
    def login(cls,
              email,
              password,
              server="https://my.farm.bot"):
        """
        We reccomend that users store tokens rather than passwords.
        """
        token = FarmbotToken.download_token(email=email,
                                            password=password,
                                            server=server)
        return Farmbot(token)

    def __init__(self, raw_token):
        token = FarmbotToken(raw_token)
        self.username = token.bot
        self.password = token.jwt
        self.hostname = token.mqtt
        self.device_id = token.sub
        self._handler = StubHandler()

        self._connection = FarmbotConnection(self)
        self.state = empty_state()

    def connect(self, handler):
        """
        Attempt to connect to the MQTT broker.
        """
        self._handler = handler
        self._connection.start_connection()

    def position(self):
        """
        Convinence method to return the bot's current location
        as an (x, y, z) tuple.
        """
        position = self.state["location_data"]["position"]
        x = position["x"] or -0.0
        y = position["y"] or -0.0
        z = position["z"] or -0.0
        return (x, y, z)

    def _do_cs(self, kind, args, body=[]):
        """
        This is a private helper that wraps CeleryScript in
        an `rpc` node and sends it to the device over MQTT.
        """
        return self._connection.send_rpc({
            "kind": kind,
            "args": args,
            "body": body
        })

    def move_absolute(self, x, y, z, speed=100.0):
        """
        Move to an absolute XYZ coordinate at a speed percentage (default speed: 100%).
        """
        return self._do_cs("move_absolute", {
            "location": {"kind": "coordinate", "args": {"x": x, "y": y, "z": z, }},
            "speed": speed,
            "offset": {"kind": "coordinate", "args": zero_xyz}
        })

    def send_message(self, msg, type="info"):
        """
        Send a log message.
        """
        return self._do_cs("send_message",
                           {"message": msg, "message_type": type, })

    def emergency_lock(self):
        """
        Perform an emergency stop, thereby preventing any
        motor movement until `emergency_unlock()` is called.
        """
        return self._do_cs("emergency_lock", {})

    def emergency_unlock(self):
        """
        Unlock the Farmduino, allowing movement of previously
        locked motors.
        """
        return self._do_cs("emergency_unlock", {})

    def find_home(self):
        """
        Find the home (0) position for all axes.
        """
        return self._do_cs("find_home", {"speed": 100, "axis": "all"})

    def find_length(self, axis="all"):
        """
        Move to the end of each axis until a stall is detected,
        then set that distance as the maximum length.
        """
        return self._do_cs("calibrate", {"axis": axis})

    def flash_farmduino(self, package="farmduino"):
        """
        Flash microcontroller firmware. `package` is one of
        the following values: "arduino", "express_k10",
        "farmduino_k14", "farmduino"
        """
        return self._do_cs("flash_firmware", {"package": package})

    def go_to_home(self, axis="all", speed=100):
        """
        Move to the home position for a given axis at a
        particular speed.
        """
        return self._do_cs("home", {"speed": speed, "axis": axis})

    def move_relative(self, x, y, z, speed=100):
        """
        Move to a relative XYZ offset from the device's current
        position at a speed percentage (default speed: 100%).
        """
        return self._do_cs("move_relative",
                           {"x": x, "y": y, "z": z, "speed": speed})

    def power_off(self):
        """
        Deactivate FarmBot OS completely (shutdown). Useful
        before unplugging the power.
        """
        return self._do_cs("power_off", {})

    def read_status(self):
        """
        Read the status of the bot. Should not be needed unless you are first
        logging in to the device, since the device pushes new states out on
        every update.
        """
        self._do_cs("read_status", {})

    def reboot(self):
        """
        Restart FarmBot OS.
        """
        return self._do_cs("reboot", {"package": "farmbot_os"})

    def reboot_farmduino(self):
        """
        Reinitialize the FarmBot microcontroller firmware.
        """
        return self._do_cs("reboot", {"package": "arduino_firmware"})

    def factory_reset(self):
        """
        THIS WILL RESET THE SD CARD, deleting all non-factory data!
        * Be careful!! *
        """
        return self._do_cs("factory_reset", {"package": "farmbot_os"})

    def sync(self):
        """
        Download/apply all of the latest FarmBot API JSON resources (plants,
        account info, etc.) to the device.
        """
        return self._do_cs("sync", {})

    def take_photo(self):
        """
        Snap a photo and send to the API for post processing.
        """
        return self._do_cs("take_photo", {})

    def toggle_pin(self, pin_number):
        """
        Reverse the value of a digital pin.
        """
        return self._do_cs("toggle_pin", {"pin_number": pin_number})

    def update_farmbot_os(self):
        return self._do_cs("check_updates", {"package": "farmbot_os"})

    def read_pin(self, pin_number, pin_mode="digital"):
        """
        Read a pin
        """
        modes = {"digital": 0, "analog": 1}
        args = {
            "label": "pin" + str(pin_number),
            "pin_mode": modes[pin_mode] or (modes["digital"]),
            "pin_number": pin_number
        }
        return self._do_cs("read_pin", args)

    def write_pin(self, pin_number, pin_value, pin_mode="digital"):
        """
        Write to a pin
        """
        modes = {"digital": 0, "analog": 1}
        args = {
            "pin_mode": modes[pin_mode] or (modes["digital"]),
            "pin_number": pin_number,
            "pin_value": pin_value
        }
        return self._do_cs("write_pin", args)

    def set_servo_angle(self, pin_number, angle):
        """
        Set pins 4 or 5 to a value between 0 and 180.
        """
        return self._do_cs("set_servo_angle",
                           {"pin_number": pin_number, "pin_value": angle})



from farmbot import Farmbot, FarmbotToken

# Before we begin, we must download an access token from the
# API. To avoid copy/pasting passwords, it is best to create
# an access token and then store that token securely:
raw_token = FarmbotToken.download_token("test@example.com",
                                        "password",
                                        "https://my.farm.bot")

# This token is then passed to the Farmbot constructor:
fb = Farmbot(raw_token)

# If you are just doing testing, such as local development,
# it is possible to skip token creation and login with email
# and password. This is not recommended for production devices:
# fb = Farmbot.login(email="em@i.l",
#                    password="pass",
#                    server="https://my.farm.bot")

# The next step is to call fb.connect(), but we are not ready
# to do that yet. Before we can call connect(), we must
# create a "handler" object. FarmBot control is event-based
# and the handler object is responsible for integrating all
# of those events into a custom application.
#
# At a minimum, the handler must respond to the following
# methods:
#     on_connect(self, bot: Farmbot, client: Mqtt) -> None
#     on_change(self, bot: Farmbot, state: Dict[Any, Any]) -> None
#     on_log(self, _bot: Farmbot, log: Dict[Any, Any]) -> None
#     on_error(self, _bot: Farmbot, _response: ErrorResponse) -> None
#     on_response(self, _bot: Farmbot, _response: OkResponse) -> None
#
# FarmBotPy will call the appropriate method whenever an event
# is triggered. For example, the method `on_log` will be
# called with the last log message every time a new log
# message is created.


class MyHandler:
    # The `on_connect` event is called whenever the device
    # connects to the MQTT server. You can place initialization
    # logic here.
    #
    # The callback is passed a FarmBot instance, plus an MQTT
    # client object (see Paho MQTT docs to learn more).
    def on_connect(self, bot, mqtt_client):
        # Once the bot is connected, we can send RPC commands.
        # Every RPC command returns a unique, random request
        # ID. Later on, we can use this ID to track our commands
        # as they succeed/fail (via `on_response` / `on_error`
        # callbacks):

        request_id1 = bot.move_absolute(x=10, y=20, z=30)
        # => "c580-6c-11-94-130002"
        print("MOVE_ABS REQUEST ID: " + request_id1)

        request_id2 = bot.send_message("Hello, world!")
        # => "2000-31-49-11-c6085c"
        print("SEND_MESSAGE REQUEST ID: " + request_id2)

    def on_change(self, bot, state):
        # The `on_change` event is most frequently triggered
        # event. It is called any time the device's internal
        # state changes. Example: Updating X/Y/Z position as
        # the device moves across the garden.
        # The bot maintains all this state in a single JSON
        # object that is broadcast over MQTT constantly.
        # It is a very large object, so we are printing it
        # only as an example.
        print("NEW BOT STATE TREE AVAILABLE:")
        print(state)
        # Since the state tree is very large, we offer
        # convenience helpers such as `bot.position()`,
        # which returns an (x, y, z) tuple of the device's
        # last known position:
        print("Current position: (%.2f, %.2f, %.2f)" % bot.position())
        # A less convenient method would be to access the state
        # tree directly:
        pos = state["location_data"]["position"]
        xyz = (pos["x"], pos["y"], pos["z"])
        print("Same information as before: " + str(xyz))

    # The `on_log` event fires every time a new log is created.
    # The callback receives a FarmBot instance, plus a JSON
    # log object. The most useful piece of information is the
    # `message` attribute, though other attributes do exist.
    def on_log(self, bot, log):
        print("New message from FarmBot: " + log['message'])

    # When a response succeeds, the `on_response` callback
    # fires. This callback is passed a FarmBot object, as well
    # as a `response` object. The most important part of the
    # `response` is `response.id`. This `id` will match the
    # original request ID, which is useful for cross-checking
    # pending operations.
    def on_response(self, bot, response):
        print("ID of successful request: " + response.id)

    # If an RPC request fails (example: stalled motors, firmware
    # timeout, etc..), the `on_error` callback is called.
    # The callback receives a FarmBot object, plus an
    # ErrorResponse object.
    def on_error(self, bot, response):
        # Remember the unique ID that was returned when we
        # called `move_absolute()` earlier? We can cross-check
        # the ID by calling `response.id`:
        print("ID of failed request: " + response.id)
        # We can also retrieve a list of error message(s) by
        # calling response.errors:
        print("Reason(s) for failure: " + str(response.errors))


# Now that we have a handler class to use, let's create an
# instance of that handler and `connect()` it to the FarmBot:
handler = MyHandler()

# Keep in mind that `connect()` will block the current thread.
# If you require parallel operations, consider using a background
# thread or a worker process.
fb.connect(handler)
print("This line will not execute. `connect()` is a blocking call.")



from setuptools import setup

with open("README.md") as fh:
    long_description = fh.read()

with open("VERSION") as vers_fh:
    version = vers_fh.read()

setup(
    name="farmbot",
    version=version,
    description="Official FarmBot RPC wrapper library for Python.",
    py_modules=["farmbot"],
    package_dir={
        "": "farmbot"
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "paho-mqtt >= 1.5",
    ],
    extras_require={
        "dev": [
            "pytest>=6.2"
        ]
    },
    url="https://github.com/farmbot-labs/farmbot-py",
    author="FarmBot, Inc.",
    author_email="contact@farmbot.io"
)
